#include "trucs_sockets.h"
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#include <string>
using std::string;

#include <Windows.h>

class Mutex
{
   HANDLE mutex_;
public:
   Mutex() throw()
      : mutex_(CreateMutex(0, FALSE, 0))
   {
   }
   ~Mutex() throw()
      { CloseHandle(mutex_); }
   void obtenir() const volatile throw()
      { WaitForSingleObject(mutex_, INFINITE); }
   void relacher() const volatile throw()
      { ReleaseMutex(mutex_); }
};

class Autoverrou
{
   const volatile Mutex &m_;
public:
   Autoverrou(const volatile Mutex &m)
      : m_(m)
   {
      m_.obtenir();
   }
   ~Autoverrou() throw()
      { m_.relacher(); }
};

class zone_transit
{
   string texte_;
   Mutex mutex_;
public:
   void ajouter(const string &s) volatile
   {
      Autoverrou av(mutex_);
      const_cast<zone_transit&>(*this).ajouter(s);
   }
   void ajouter(const string &s)
      { texte_ += s; }
   string extraire() volatile
   {
      Autoverrou av(mutex_);
      return const_cast<zone_transit&>(*this).extraire();
   }
   string extraire()
   {
      string s = texte_;
      texte_.clear();
      return s;
   }
};

struct Relais
{
   SOCKET sck;
   bool &fini;
   zone_transit sortie;
   zone_transit entree;
   Relais(SOCKET sck, bool &fini)
      : sck(sck), fini(fini)
   {
   }
   ~Relais() throw()
   {
      FermerSocket(sck);
   }
};

unsigned long __stdcall lire_clavier(void *p)
{
   using std::getline;
   using std::cin;
   Relais &r = *reinterpret_cast<Relais*>(p);
   for(string s; !r.fini && getline(cin, s); )
   {
      if (s != ".")
         r.sortie.ajouter(s + '\n');
      else
         r.fini = true;
   }
   FermerSocket(r.sck);
   r.fini = true;
   return 0UL;
}

unsigned long __stdcall envoyer_socket(void *p)
{
   Relais &r = *reinterpret_cast<Relais*>(p);
   while (!r.fini)
      try
      {
         string s = r.sortie.extraire();
         if (!s.empty())
            Envoyer(r.sck, s, '\0');
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fini = true;
      }
   return 0UL;
}

unsigned long __stdcall consommer_socket(void *p)
{
   Relais &r = *reinterpret_cast<Relais*>(p);
   while (!r.fini)
      try
      {
         string s = Recevoir(r.sck, '\0');
         if (!s.empty())
            r.entree.ajouter(s);
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fini = true;
      }
   return 0UL;
}
unsigned long __stdcall ecrire_console(void *p)
{
   using std::flush;
   Relais &r = *reinterpret_cast<Relais*>(p);
   while (!r.fini)
      try
      {
         string s = r.entree.extraire();
         if (!s.empty())
            cout << s << flush;
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fini = true;
      }
   return 0UL;
}
#include <algorithm>
int main()
{
   const string ADRESSE_DU_SERVEUR = "127.0.0.1";
   const short PORT_DU_SERVEUR = normaliser(static_cast<short>(4321));
   const string MESSAGE = "a marche!";
   const char DELIMITEUR_FIN = '#';
   // Initialisation des sockets pour le client
   ChargerSockets();
   // Obtention du socket client
   socket_t client = CreerSocket();
   // Tentative de connexion au serveur
   Connecter(client, ADRESSE_DU_SERVEUR, PORT_DU_SERVEUR);

   bool fini = false;
   Relais r(client, fini);
   HANDLE h[] =
   {
      CreateThread(0, 0, lire_clavier, &r, 0, 0),
      CreateThread(0, 0, envoyer_socket, &r, 0, 0),
      CreateThread(0, 0, consommer_socket, &r, 0, 0),
      CreateThread(0, 0, ecrire_console, &r, 0, 0)
   };

   enum { N = sizeof(h) / sizeof(*h) };
   WaitForMultipleObjects(N, h, TRUE, INFINITE);
   std::for_each(std::begin(h), std::end(h), [](HANDLE hdl) {
      CloseHandle(hdl);
   });
   DechargerSockets();
}

